using System;
using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.Events;
using NaughtyAttributes;

public class StepsListScript : MonoBehaviour
{
    // Const strings
    private const string LOG_START_LIST = "Starting list coroutine";

    [Header("Disclaimer")]
    [InfoBox("This script is part of the StepsLink system. It loads represent a list of steps do launch in order.", EInfoBoxType.Normal)]

    [HorizontalLine(color: EColor.Blue)]

    [Header("Callbacks")]
    [InfoBox("Callbacks are called before starting the first step, or after ending the last.", EInfoBoxType.Normal)]
    /// <summary>
    /// If true, run the callbacks when the list starts
    /// </summary>
    [Tooltip("If true, run the callbacks when the list starts")]
    public bool useOnStart = false;

    /// <summary>
    /// Event called when the list starts
    /// </summary>
    [ShowIf("useOnStart")]
    [Tooltip("Event called when the list starts")]
    public UnityEvent OnStart;

    /// <summary>
    /// If true, run the callbacks when the list ends
    /// </summary>
    [Tooltip("If true, run the callbacks when the list ends")]
    public bool useOnEnd = false;

    /// <summary>
    /// Event called when the list ends
    /// </summary>
    [Tooltip("Event called when the list ends")]
    [ShowIf("useOnEnd")]
    public UnityEvent OnEnd;

    /// <summary>
    /// If true, once the OnEnd callback are called, a delay is started, this allows to
    /// run another set of callback after these
    /// </summary>
    [Tooltip("If true, once the OnEnd callback are called, a delay is started, this allows to " +
        "run another set of callback after these")]
    public bool useDelayedEnd = false;

    /// <summary>
    ///  Delay after the list ends, used to wait for delayed effects
    /// </summary>
    [ShowIf("useDelayedEnd")]
    [MinValue(0.0f), MaxValue(60)]
    [Tooltip("Delay after the list ends, used to wait for delayed effects, runned before the OnEndPostDelay")]
    public float endTimeDurationSeconds = 5f;

    /// <summary>
    /// Event called after the list ends, after a delay
    /// </summary>
    [ShowIf("useDelayedEnd")]
    [Tooltip("Event called after the list ends, after a delay")]
    public UnityEvent OnEndPostDelay;

    [HorizontalLine(color: EColor.Blue)]

    /// <summary>
    /// If true, the list of steps will be loaded from the children of this gameObject
    /// else, the list of steps must be added manually
    /// </summary>
    [Header("How to load steps?")]
    [InfoBox("To be loaded, a step must have StepScript component", EInfoBoxType.Normal)]
    [Tooltip("If true, the list of steps will be loaded from the children of this gameObject")]
    public bool useChildrenAsSteps = true;

    /// <summary>
    /// List of steps to run, a step must have the StepScript component attached
    /// </summary>
    [InfoBox("Childen of this list will be ignored. All steps must be added manually", EInfoBoxType.Normal)]
    [Tooltip("List of steps to run, manually linked from any position")]
    [HideIf("useChildrenAsSteps")]
    public List<StepScript> stepsList = new();

    /// <summary>
    /// True if the list is completed, after the StartSteps() coroutine is completed
    /// </summary>
    private bool isCompleted;

    /// <summary>
    /// True if the list is completed, after the StartSteps() coroutine is completed
    /// </summary>
    internal bool IsCompleted { get => isCompleted; }

    /// <summary>
    /// True if the list is initialized, after the InitListSteps() coroutine is completed
    /// <see cref="InitListSteps"/>"
    /// </summary>
    private bool initialized = false;

    /// <summary>
    /// True if the list is initialized, after the InitListSteps() coroutine is completed
    /// </summary>
    public bool Initialized { get => initialized; }

    /// <summary>
    /// Number of steps in the list
    /// </summary>
    private int StepsCount = 0;

    /// <summary>
    /// Number of steps in the list
    /// </summary>
    public int GetStepsCount { get => StepsCount; }

    /// <summary>
    /// Initialize the list of steps. This is called by the StepOrchestrator to initialize the list
    /// </summary>
    /// <returns></returns>
    /// <exception cref="Exception"></exception>
    public IEnumerator InitListSteps()
    {
        isCompleted = false;

        // get all the children ojjects that have the Step script attached
        if (useChildrenAsSteps)
        {
            stepsList.Clear();

            foreach (Transform child in transform)
            {
                if (child.GetComponent<StepScript>() != null)
                {
                    stepsList.Add(child.GetComponent<StepScript>());
                }

                if (stepsList.Count == 0)
                {
                    Debug.LogWarning("No steps found in the children of the list: " + name);
                }
            }
        }
        else if (stepsList.Count == 0)
        {
            Debug.LogWarning("No steps found in the list: " + name);
        }
        else if (stepsList.Count > 0)
        {
            foreach (StepScript step in stepsList)
            {
                Debug.Log("Loading manual list and filtering steps in the list: " + name);

                // check if there is a StepScript component attached to every stel added
                if (step.GetComponent<StepScript>() == null)
                {
                    throw new Exception("The step: " + step.name + " has not the StepScript attached");
                }
            }
        }

        StepsCount = stepsList.Count;

        initialized = true;

        yield return null;
    }

    /// <summary>
    /// Start the list of steps. This is called by the StepOrchestrator when the list is ready
    /// </summary>
    /// <returns></returns>
    public IEnumerator StartSteps()
    {
        Debug.Log(LOG_START_LIST + ": " + name);

        if (useOnStart)
        {
            OnStart.Invoke();
        }

        foreach (StepScript step in stepsList)
        {
            // wait for the step to complete
            yield return StartCoroutine(step.StepRun());
        }

        if (useOnEnd)
        {
            OnEnd.Invoke();
        }

        if (useDelayedEnd)
        {
            yield return new WaitForSeconds(endTimeDurationSeconds);
            OnEndPostDelay.Invoke();
        }

        Debug.Log("List: " + name + " completed");


        isCompleted = true;
    }
}
